
function WebRtcClient(uid, signalingUrl) {
    var that = this;
    this.socket = io(signalingUrl);
    this.id = {uid: uid, sid: null};
    this.peer = null;
    this.peerConnection = null;
    this.iceServers = null;
    this.iceServersConf = null;
    this.localStream = null;
    this.queue = [];
    const offerOptions = {
        offerToReceiveAudio: 1,
        offerToReceiveVideo: 1
    };
    this.inited = false;

    this.onPeerConnecting = function(peerInfo) {

    };

    this.onPeerConnected = function(peerInfo) {

    };

    this.onRemoteStream = function (remoteStreams) {

    };

    this.onPeerDisconnected = function(peerInfo) {

    };

    this.onPrepareLocalStream = function(onStreamReady) {

    };

    this.initPeer = function initPeer(peerInfo, pcId, ices, onReady) {
        this.onPrepareLocalStream(function(stream) {
            that.localStream = stream;
            console.log("got local Stream " + stream);
            that.initPeerInternal(peerInfo, pcId, ices, onReady);
        });
    };
    this.initPeerInternal = function(peerInfo, pcId, ices, onReady) {
        console.log("init " + " pcid:" + pcId + " peer: " + JSON.stringify(peerInfo));
        this.peer = peerInfo;
        this.pcId = pcId;
        this.iceServers = ices;
        this.iceServersConf = serversToServersConf(ices)
        var configuration = {'iceServers': this.iceServersConf};
        this.onPeerConnecting(peerInfo);
        this.peerConnection = new RTCPeerConnection(configuration);
        this.localStream.getVideoTracks().forEach(track => console.log("video track:" + track));
        this.localStream.getAudioTracks().forEach(track => console.log("audio track:" + track));
        this.localStream.getTracks().forEach(track => that.peerConnection.addTrack(track, that.localStream));
        console.log('Added local stream to peerConnection');
        this.peerConnection.addEventListener('icecandidate', event => {
            console.log("peerConnection event " + this.peerConnection.connectionState);
            if (event.candidate) {
                that.socket.emit("message", candidateMsg(that.peer, that.pcId, event.candidate));
            }
        });
        console.log("1");
        this.peerConnection.addEventListener('connectionstatechange', event => {
            if (that.peerConnection == null) {
                console.log("connectionstatechange " + null);
                return;
            }
            console.log("connectionstatechange " + JSON.stringify(event) + " state:" + that.peerConnection.connectionState);
            if (that.peerConnection.connectionState === 'connected') {
                // Peers connected!
                console.log("peer connected");
                that.onPeerConnected(that.peer);
            } else if (that.peerConnection.connectionState === 'disconnected') {
                console.log("peer lost");
                that.onPeerDisconnected(that.peer);
            }
        });
        this.peerConnection.ontrack = function(event) {
            console.log("ontrack");
            console.log('pc2 received remote stream', JSON.stringify(event.streams));
            that.onRemoteStream(event.streams);
            // if (remoteVideo.srcObject !== event.streams[0]) {
                // remoteVideo.srcObject = event.streams[0];
            // }

        };
        console.log("3 " + onReady);
        this.inited = true;
        onReady();
        console.log("inited queue:" + this.queue.length);
        for (var i in this.queue)
            this.queue[i]();
    };
    this.socket.on('connect', function() {
        console.log('socket connect');
    });
    this.socket.on('disconnect', function() {
        console.log('socket disconnect');
    });
    this.socket.on('error', function() {
        console.log('socket error');
    });
    this.socket.on('sid', function(data){
        that.id.sid = data;
        console.log("connected sid:" + that.id.sid);
        var msg = new Object();
        msg.uid = that.id.uid;
        that.socket.emit("enterPairingMode", msg);
    });
    this.socket.on('init', async function(data){
        console.log("on init ");
        that.initPeer(data.peer, guid(), data.iceServers, async function() {
            console.log("offerOptions " + JSON.stringify(offerOptions));
            const offer = await that.peerConnection.createOffer(offerOptions);
            await that.peerConnection.setLocalDescription(offer);
            console.log("offer:" + offer);
            console.log("offer.type:" + offer.type);
            // console.log("offer.sdp:" + offer.sdp);
            that.socket.emit("message", offerMsg(that.peer, that.pcId, that.iceServers, offer));
        });
    });
    this.socket.on('message', async function(data){
        console.log("on msg type: " + data.type + " from: " + JSON.stringify(data.from));
        if (data.type == "offer") {
            console.log("0");
            that.initPeer(data.from, data.payload.pcid, data.payload.iceServers, async function() {
                console.log("10");
                console.log("offer msg payload:" + JSON.stringify(data.payload));
                that.peerConnection.setRemoteDescription(new RTCSessionDescription(data.payload));
                console.log("anwserOptions " + JSON.stringify(offerOptions));
                const answer = await that.peerConnection.createAnswer(offerOptions);
                await that.peerConnection.setLocalDescription(answer);
                console.log("answer to " + JSON.stringify(that.peer) + " answer: " + JSON.stringify(answer));
                that.socket.emit("message", answerMsg(that.peer, that.pcId, answer));
            });
        } else if (data.type == "answer") {
            console.log("answer msg " + JSON.stringify(data.payload));
            const remoteDesc = new RTCSessionDescription(data.payload);
            await that.peerConnection.setRemoteDescription(remoteDesc);

        } else if (data.type == "candidate") {
            console.log("candidate msg ");
            that.doAfterInit(function() {
                var candidate = new Object();
    
                // console.log("remote candidate payload", JSON.stringify(data.payload));
                candidate.sdpMLineIndex = data.payload.label;
                candidate.sdpMid = data.payload.id;
                candidate.candidate = data.payload.candidate;
                // console.log("remote candidate payload", JSON.stringify(candidate));
                that.peerConnection.addIceCandidate(new RTCIceCandidate(candidate));

            });
        }


    });

    this.socket.on('fail', function(data){
        console.log("fail uid:" + data.peerId.uid + " reason:" + data.reason);
        if (that.peer == null || data.peerId.uid != that.peer.id.uid) {
            console.log ("fail uid not match current: " + JSON.stringify(that.peer));
        }
        that.stopCurrentPeer();
    });

    this.stopCurrentPeer = function() {
        console.log("stopCurrentPeer");
        if(this.peerConnection != null) this.peerConnection.close();
        this.iceServersConf = null;
        this.iceServers = null;
        this.peerConnection = null;
        this.localStream = null;
        this.queue = [];
        this.inited = false;
        if (this.peer != null) {
            this.peer = null;
            this.onPeerDisconnected();
        }
    };

    this.next = function() {
        console.log("next cpeer:" + this.peer);
        if (this.peer == null)
            return;
        console.log("real next");
        var msg = new Object();
        msg.peerId = this.peer.id;
        this.stopCurrentPeer();
        msg.reasong = "next";
        this.socket.emit("next", msg);
    };

    this.dispose = function() {
        console.log("WebRtcClient  dispose");
        this.socket.close();
        this.stopCurrentPeer();
    };

    this.doAfterInit = function(f) {
        if (this.inited) {
            f();
        } else  {
            this.queue.push(f);
        }
    }
}

function serversToServersConf(servers) {
    console.log("serversToServersConf " + JSON.stringify(servers));
    var confs = [];
    for (var i in servers) {
        var server = servers[i];
        var serverConf = new Object();
        serverConf.urls = [server.uri];
        serverConf.username = server.username;
        serverConf.credential = server.password;
        console.log("ice:" + serverConf);
        confs.push(serverConf);
    }
    return confs;
}


function offerMsg(peer, pcId, iceServers, offer) {
    var payload = new Object();
    var type = "offer";
    payload.type = type;
    // console.log(offer.sdp);
    payload.sdp = offer.sdp;
    payload.iceServers = iceServers;
    return msg(peer, type, pcId, payload);
}
function answerMsg(peer, pcId, answer) {
    var payload = new Object();
    var type = "answer";
    payload.type = type;
    payload.sdp = answer.sdp;
    return msg(peer, type, pcId, payload);
}
function candidateMsg(peer, pcId, candidate) {
    console.log("local candidate origin", JSON.stringify(candidate));
    var payload = new Object()
    var type = "candidate";
    payload.type = type;
    payload.label = candidate.sdpMLineIndex;
    payload.id = candidate.sdpMid;
    payload.candidate = candidate.candidate;
    console.log("local candidate payload", JSON.stringify(payload));
    return msg(peer, type, pcId, payload);
}
function msg(to, type, pcId, payload) {
    var msg = new Object();
    msg.to = to;
    msg.type = type;
    payload.pcid = pcId;
    msg.payload = payload;
    console.log("send: " + JSON.stringify(msg));
    return msg;
}

function guid() {
    function S4() {
       return (((1+Math.random())*0x10000)|0).toString(16).substring(1);
    }
    return (S4()+S4()+"-"+S4()+"-"+S4()+"-"+S4()+"-"+S4()+S4()+S4());
}

// module.exports = {
//     WebRtcClient,
// }